import collections
import multiprocessing as mp
import os
import pathlib
import threading
import traceback

import cloudpickle
import psutil

from . import utils


class Context:

  def __init__(self):
    self.initfns = []
    self.resolver = None
    self.errfile = None
    self.interval = 20
    self.clientkw = {}
    self.serverkw = {}
    self.done = threading.Event()
    self.watcher = None
    self.children = collections.defaultdict(list)
    self.mp = mp.get_context()
    self.printlock = self.mp.Lock()

  def options(self):
    return {
        'resolver': self.resolver and cloudpickle.dumps(self.resolver),
        'errfile': self.errfile,
        'interval': self.interval,
        'initfns': self.initfns,
        'clientkw': self.clientkw,
        'serverkw': self.serverkw,
    }

  def setup(
      self,
      resolver=None,
      errfile=None,
      interval=None,
      initfns=None,
      clientkw=None,
      serverkw=None,
      host=None,
      ipv6=None,
  ):

    if resolver:
      if isinstance(resolver, bytes):
        resolver = cloudpickle.loads(resolver)
      assert callable(resolver)
      self.resolver = resolver

    if errfile:
      if isinstance(errfile, str):
        errfile = pathlib.Path(errfile)
      assert hasattr(errfile, 'exists') and hasattr(errfile, 'write_text')
      self.errfile = errfile

    if interval:
      assert isinstance(interval, (int, float))
      self.interval = interval

    if initfns:
      self.initfns = []
      for fn in initfns:
        self.initfn(fn, call_now=True)

    if clientkw is not None:
      assert isinstance(clientkw, dict)
      self.clientkw = clientkw

    if serverkw is not None:
      assert isinstance(serverkw, dict)
      self.serverkw = serverkw

    if host is not None:
      self.serverkw['host'] = host

    if ipv6 is not None:
      self.clientkw['ipv6'] = ipv6
      self.serverkw['ipv6'] = ipv6

    if self.errfile and not self.watcher:
      self.watcher = threading.Thread(target=self._watcher, daemon=True)
      self.watcher.start()

  def initfn(self, fn, call_now=True):
    if isinstance(fn, bytes):
      pkl, fn = fn, cloudpickle.loads(fn)
    else:
      pkl, fn = cloudpickle.dumps(fn), fn
    self.initfns.append(pkl)
    call_now and fn()

  def print(self, name, *args, color=None):
    assert args
    if color:
      style = utils.style(color=color)
      reset = utils.style(reset=True)
    else:
      style, reset = '', ''
    with self.printlock:
      print(style + f'[{name}]' + reset, *args)

  def error(self, e, name=None):
    typ, tb = type(e), e.__traceback__
    summary = list(traceback.format_exception_only(typ, e))[0].strip('\n')
    long = ''.join(traceback.format_exception(typ, e, tb)).strip('\n')
    message = f"Error in '{name}' ({summary}):\n{long}"
    with self.printlock:
      style = utils.style(color='red')
      reset = utils.style(reset=True)
      print(style + '\n---\n' + message + reset)
    if self.errfile:
      self.errfile.write_text(message)
      print(f'Wrote errorfile: {self.errfile}')

  def shutdown(self, exitcode):
    # This kills the process tree forcefully to prevent hangs but results in
    # leaked semaphore warnings. However, the leaked objects are still cleaned
    # up by the resource tracker process of Python's multiprocessing module.
    utils.kill_procs(psutil.Process().children(recursive=True))
    os._exit(exitcode)

  def close(self):
    self.done.set()
    if self.watcher:
      self.watcher.join()

  def add_child(self, worker):
    ident = threading.get_ident()
    if hasattr(worker, 'ident'):
      assert worker.ident != ident
    self.children[ident].append(worker)

  def get_children(self, ident=None):
    if ident is None:
      ident = threading.get_ident()
    return self.children[ident]

  def _watcher(self):
    while True:
      if self.errfile and self.errfile.exists():
        print(f'Shutting down due to error file: {self.errfile}')
        self.shutdown(2)
        break
      if self.done.wait(self.interval):
        break


context = Context()


def initfn(fn, call_now=True):
  context.initfn(fn, call_now)


def setup(**kwargs):
  context.setup(**kwargs)


def reset():
  global context
  context.close()
  context = Context()



import pathlib
import time

import pytest
import portal


class TestErrfile:

  def test_file(self, tmpdir):
    errfile = pathlib.Path(tmpdir) / 'error'

    def fn():
      portal.setup(errfile=errfile, interval=0.1)
      try:
        raise ValueError
      except Exception as e:
        portal.context.error(e, 'worker')

    portal.Process(fn, start=True).join()
    content = errfile.read_text()
    assert "Error in 'worker' (ValueError):" == content.split('\n')[0]
    assert 'Traceback (most recent call last)' in content
    assert 'line' in content
    assert 'in fn' in content

  @pytest.mark.parametrize('repeat', range(3))
  def test_sibling_procs(self, tmpdir, repeat):
    errfile = pathlib.Path(tmpdir) / 'error'
    ready = portal.context.mp.Semaphore(0)

    def fn1(ready, errfile):
      portal.setup(errfile=errfile, interval=0.1)
      ready.release()
      raise ValueError('reason')

    def fn2(ready, errfile):
      portal.setup(errfile=errfile, interval=0.1)
      ready.release()
      while True:
        time.sleep(0.1)

    worker1 = portal.Process(fn1, ready, errfile, name='worker1', start=True)
    worker2 = portal.Process(fn2, ready, errfile, name='worker2', start=True)
    ready.acquire()
    ready.acquire()
    worker1.join()
    worker2.join()
    content = errfile.read_text()
    first_line = content.split('\n')[0]
    assert "Error in 'worker1' (ValueError: reason):" == first_line
    assert not worker1.running
    assert not worker2.running
    assert worker1.exitcode == 1
    assert worker2.exitcode == 2

  @pytest.mark.parametrize('repeat', range(3))
  def test_nested_procs(self, tmpdir, repeat):
    errfile = pathlib.Path(tmpdir) / 'error'
    ready = portal.context.mp.Semaphore(0)

    def hang():
      while True:
        time.sleep(0.1)

    def outer(ready, errfile):
      portal.setup(errfile=errfile, interval=0.1)
      portal.Process(inner, ready, errfile, name='inner', start=True)
      portal.Thread(hang, start=True)
      portal.Process(hang, start=True)
      ready.release()
      hang()

    def inner(ready, errfile):
      portal.setup(errfile=errfile, interval=0.1)
      portal.Thread(hang, start=True)
      portal.Process(hang, start=True)
      ready.release()
      raise ValueError('reason')

    worker = portal.Process(outer, ready, errfile, name='outer', start=True)
    ready.acquire()
    ready.acquire()
    worker.join()
    content = errfile.read_text()
    assert "Error in 'inner' (ValueError: reason):" == content.split('\n')[0]
    assert not worker.running



import struct

import msgpack
import numpy as np

from . import sharray


def pack(data):
  leaves, treedef = tree_flatten(data)
  specs, buffers = [], []
  for value in leaves:
    if value is None:
      specs.append(['none'])
      buffers.append(b'\x00')
    elif isinstance(value, str):
      specs.append(['utf8'])
      buffers.append(value.encode('utf-8'))
    elif isinstance(value, (bytes, bytearray, memoryview)):
      if isinstance(value, memoryview):
        value = value.cast('c')
        assert value.c_contiguous
      specs.append(['bytes'])
      buffers.append(value)
    elif isinstance(value, (np.ndarray, np.generic, int, float)):
      value = np.asarray(value)
      if value.dtype == object:
        raise TypeError(data)
      assert value.data.c_contiguous, (
          "Array is not contiguous in memory. Use " +
          "np.asarray(arr, order='C') before passing the data into pack().")
      specs.append(['array', value.shape, value.dtype.str])
      buffers.append(value.data.cast('c'))
    elif isinstance(value, sharray.SharedArray):
      specs.append(['sharray', *value.__getstate__()])
      buffers.append(b'\x00')
    else:
      raise NotImplementedError(type(value))
  buffers = [msgpack.packb(treedef), msgpack.packb(specs), *buffers]
  length = len(buffers).to_bytes(8, 'little', signed=False)
  sizes = struct.pack('<' + ('Q' * len(buffers)), *[len(x) for x in buffers])
  buffers = [length, sizes, *buffers]
  return buffers


def unpack(buffer):
  length = int.from_bytes(buffer[:8], 'little', signed=False)
  buffer = buffer[8:]
  sizes = struct.unpack('<' + ('Q' * length), buffer[:8 * length])
  buffer = buffer[8 * length:]
  limits = np.cumsum(sizes)
  buffers = [buffer[i: j] for i, j in zip([0, *limits[:-1]], limits)]
  treedef, specs, *buffers = buffers
  treedef = msgpack.unpackb(treedef)
  specs = msgpack.unpackb(specs)
  leaves = []
  for spec, buffer in zip(specs, buffers):
    if spec[0] == 'none':
      assert buffer == b'\x00'
      leaves.append(None)
    elif spec[0] == 'utf8':
      leaves.append(bytes(buffer).decode('utf-8'))
    elif spec[0] == 'bytes':
      leaves.append(buffer)
    elif spec[0] == 'array':
      shape, dtype = spec[1:]
      leaves.append(np.frombuffer(buffer, dtype).reshape(shape))
    elif spec[0] == 'sharray':
      assert buffer == b'\x00'
      leaves.append(sharray.SharedArray(*spec[1:]))
    else:
      raise NotImplementedError(spec)
  data = tree_unflatten(leaves, treedef)
  return data


def tree_map(fn, *trees, isleaf=None):
  assert trees, 'Provide one or more nested Python structures'
  kw = dict(isleaf=isleaf)
  first = trees[0]
  try:
    assert all(isinstance(x, type(first)) for x in trees)
    if isleaf and isleaf(trees[0]):
      return fn(*trees)
    if isinstance(first, list):
      assert all(len(x) == len(first) for x in trees)
      return [tree_map(
          fn, *[t[i] for t in trees], **kw) for i in range(len(first))]
    if isinstance(first, tuple):
      assert all(len(x) == len(first) for x in trees)
      return tuple([tree_map(
          fn, *[t[i] for t in trees], **kw) for i in range(len(first))])
    if isinstance(first, dict):
      assert all(set(x.keys()) == set(first.keys()) for x in trees)
      return {k: tree_map(fn, *[t[k] for t in trees], **kw) for k in first}
    if hasattr(first, 'keys') and hasattr(first, 'get'):
      assert all(set(x.keys()) == set(first.keys()) for x in trees)
      return type(first)(
          {k: tree_map(fn, *[t[k] for t in trees], **kw) for k in first})
  except AssertionError:
    raise TypeError('Tree structure differs between arguments')
  return fn(*trees)


def tree_flatten(tree, isleaf=None):
  leaves = []
  tree_map(lambda x: leaves.append(x), tree, isleaf=isleaf)
  structure = tree_map(lambda x: None, tree, isleaf=isleaf)
  return tuple(leaves), structure


def tree_unflatten(leaves, structure):
  leaves = iter(tuple(leaves))
  return tree_map(lambda x: next(leaves), structure)


def tree_equals(xs, ys):
  assert type(xs) == type(ys), (type(xs), type(ys))
  if isinstance(xs, (list, tuple)):
    assert len(xs) == len(ys)
    return all(tree_equals(x, y) for x, y in zip(xs, ys))
  elif isinstance(xs, dict):
    assert xs.keys() == ys.keys()
    return all(tree_equals(xs[k], ys[k]) for k in xs.keys())
  elif isinstance(xs, np.ndarray):
    assert xs.shape == ys.shape, (xs.shape, ys.shape)
    assert xs.dtype == ys.dtype, (xs.dtype, ys.dtype)
    return (xs == ys).all()
  elif isinstance(xs, sharray.SharedArray):
    assert xs.name == ys.name, (xs.name, ys.name)
    assert xs.array.shape == ys.array.shape, (xs.array.shape, ys.array.shape)
    assert xs.array.dtype == ys.array.dtype, (xs.array.dtype, ys.array.dtype)
    return (xs == ys).all()
  else:
    return xs == ys



import threading
import time
import traceback

from . import contextlib
from . import utils


class Thread:

  """
  This thread implementation extends the standard Python thread as follows:

  1. It provide a kill() method that raises a SystemExit error inside the
  thread. This allows killing the thread from the outside.

  2. It keeps track of its subthreads and subprocesses and kills them on error
  to prevent hangs.

  3. The thread is marked as daemon, so that when the parent process dies the
  thread is killed, preventing hangs.
  """

  def __init__(self, fn, *args, name=None, start=False):
    self.fn = fn
    self.excode = None
    name = name or getattr(fn, '__name__', 'thread')
    self.thread = threading.Thread(
        target=self._wrapper, args=args, name=name, daemon=True)
    contextlib.context.add_child(self)
    self.started = False
    start and self.start()

  @property
  def name(self):
    return self.thread.name

  @property
  def ident(self):
    return self.thread.ident

  @property
  def running(self):
    if not self.started:
      return False
    return self.thread.is_alive()

  @property
  def exitcode(self):
    return self.excode

  def start(self):
    assert not self.started
    self.started = True
    self.thread.start()
    return self

  def join(self, timeout=None):
    if self.running:
      self.thread.join(timeout)
    return self

  def kill(self, timeout=3):
    utils.kill_threads(self.thread, timeout)
    return self

  def __repr__(self):
    attrs = ('name', 'ident', 'running', 'exitcode')
    attrs = [f'{k}={getattr(self, k)}' for k in attrs]
    return 'Thread(' + ', '.join(attrs) + ')'

  def _wrapper(self, *args):
    try:
      exitcode = self.fn(*args)
      exitcode = exitcode if isinstance(exitcode, int) else 0
      self.excode = exitcode
    except (SystemExit, KeyboardInterrupt) as e:
      compact = traceback.format_tb(e.__traceback__)
      compact = '\n'.join([line.split('\n', 1)[0] for line in compact])
      print(f"Killed thread '{self.name}' at:\n{compact}")
      [x.kill(0.1) for x in contextlib.context.get_children()]
      self.excode = 2
    except Exception as e:
      [x.kill(0.1) for x in contextlib.context.get_children()]
      contextlib.context.error(e, self.name)
      contextlib.context.shutdown(1)
      self.excode = 1



import atexit
import traceback

import cloudpickle
import psutil

from . import contextlib
from . import utils


class Process:

  """
  This process implementation extends the standard Python process as follows:

  1. It provides a kill() method that not only terminates the process itself
  but also all its nested child processes. This prevents subprocesses from
  lingering in the background after the Python program has ended.

  2. The process terminates its nested child processes when it encounters an
  error. This prevents lingering subprocesses on error.

  3. When the parent process encounters an error, the subprocess will be killed
  via `atexit`, preventing hangs.

  4. It inherits the context object of its parent process, which provides
  cloudpickled initializer functions for each nested child process and error
  file watching for global shutdown on error.
  """

  def __init__(self, fn, *args, name=None, start=False):
    name = name or getattr(fn, '__name__', 'process')
    fn = cloudpickle.dumps(fn)
    context = contextlib.context
    options = context.options()
    self.process = context.mp.Process(
        target=self._wrapper, name=name, args=(options, name, fn, args))
    context.add_child(self)
    self.started = False
    atexit.register(self.kill)
    start and self.start()

  @property
  def name(self):
    return self.process.name

  @property
  def pid(self):
    return self.process.pid

  @property
  def running(self):
    if not self.started:
      return False
    return self.process.is_alive()

  @property
  def exitcode(self):
    if not self.started or self.running:
      return None
    else:
      return self.process.exitcode

  def start(self):
    assert not self.started
    self.started = True
    self.process.start()
    assert self.pid is not None
    return self

  def join(self, timeout=None):
    assert self.started
    if self.running:
      self.process.join(timeout)
    return self

  def kill(self, timeout=1):
    try:
      proc = psutil.Process(self.pid)
      tree = [proc] + list(proc.children(recursive=True))
    except psutil.NoSuchProcess:
      tree = []
    self.process.terminate()
    self.process.join(timeout / 2)
    self.process.kill()
    utils.kill_procs(tree, timeout / 2)
    return self

  def __repr__(self):
    attrs = ('name', 'pid', 'running', 'exitcode')
    attrs = [f'{k}={getattr(self, k)}' for k in attrs]
    return 'Process(' + ', '.join(attrs) + ')'

  @staticmethod
  def _wrapper(options, name, fn, args):
    exitcode = 0
    try:
      contextlib.setup(**options)
      fn = cloudpickle.loads(fn)
      exitcode = fn(*args)
      exitcode = exitcode if isinstance(exitcode, int) else 0
    except (SystemExit, KeyboardInterrupt) as e:
      compact = traceback.format_tb(e.__traceback__)
      compact = '\n'.join([line.split('\n', 1)[0] for line in compact])
      print(f"Killed process '{name}' at:\n{compact}")
      exitcode = 2
    except Exception as e:
      contextlib.context.error(e, name)
      exitcode = 1
    finally:
      contextlib.context.shutdown(exitcode)



import collections
import dataclasses
import queue
import selectors
import socket

from . import buffers
from . import contextlib
from . import thread


class Connection:

  def __init__(self, sock, addr):
    self.sock = sock
    self.addr = addr
    self.recvbuf = None
    self.sendbufs = collections.deque()

  def fileno(self):
    return self.sock.fileno()


@dataclasses.dataclass
class Options:

  ipv6: bool = False
  host: str = ''
  max_msg_size: int = 4 * 1024 ** 3
  max_recv_queue: int = 4096
  max_send_queue: int = 4096
  logging: bool = True
  logging_color: str = 'blue'


class ServerSocket:

  def __init__(self, port, name='Server', **kwargs):
    if isinstance(port, str):
      port = int(port.rsplit(':', 1)[1])
    self.name = name
    self.options = Options(**{**contextlib.context.serverkw, **kwargs})
    if self.options.ipv6:
      self.sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
      self.addr = (self.options.host, port, 0, 0)
    else:
      self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      self.addr = (self.options.host, port)
    self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    # self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    self.sock.bind(self.addr)
    self.sock.setblocking(False)
    self.sock.listen()
    self.sel = selectors.DefaultSelector()
    self.sel.register(self.sock, selectors.EVENT_READ, data=None)
    self._log(f'Listening at {self.addr[0]}:{self.addr[1]}')
    self.conns = {}
    self.recvq = queue.Queue()  # [(addr, bytes)]
    self.reading = True
    self.running = True
    self.error = None
    self.thread = thread.Thread(self._loop, name=f'{name}Loop', start=True)

  @property
  def connections(self):
    return tuple(self.conns.keys())

  def recv(self, timeout=None):
    if self.error:
      raise self.error
    assert self.running
    try:
      return self.recvq.get(block=(timeout != 0), timeout=timeout)
    except queue.Empty:
      raise TimeoutError

  def send(self, addr, *data):
    if self.error:
      raise self.error
    assert self.running
    if self._numsending() > self.options.max_send_queue:
      raise RuntimeError('Too many outgoing messages enqueued')
    maxsize = self.options.max_msg_size
    try:
      self.conns[addr].sendbufs.append(
          buffers.SendBuffer(*data, maxsize=maxsize))
    except KeyError:
      self._log('Dropping message to disconnected client')

  def shutdown(self):
    self.reading = False

  def close(self, timeout=None):
    self.running = False
    self.thread.join(timeout)
    [conn.sock.close() for conn in self.conns.values()]
    self.sock.close()
    self.sel.close()

  def _loop(self):
    try:
      while self.running or self._numsending():
        writeable = []
        for key, mask in self.sel.select(timeout=0.2):
          if key.data is None and self.reading:
            assert mask & selectors.EVENT_READ
            self._accept(key.fileobj)
          elif mask & selectors.EVENT_READ and self.reading:
            self._recv(key.data)
          elif mask & selectors.EVENT_WRITE:
            writeable.append(key.data)
        for conn in writeable:
          if not conn.sendbufs:
            continue
          try:
            conn.sendbufs[0].send(conn.sock)
            if conn.sendbufs[0].done():
              conn.sendbufs.popleft()
          except BlockingIOError:
            pass
          except ConnectionResetError:
            # The client is gone but we may have buffered messages left to
            # read, so we keep the socket open until recv() fails.
            pass
    except Exception as e:
      self.error = e

  def _accept(self, sock):
    sock, addr = sock.accept()
    self._log(f'Accepted connection from {addr[0]}:{addr[1]}')
    sock.setblocking(False)
    conn = Connection(sock, addr)
    self.sel.register(
        sock, selectors.EVENT_READ | selectors.EVENT_WRITE, data=conn)
    self.conns[addr] = conn

  def _recv(self, conn):
    if not conn.recvbuf:
      conn.recvbuf = buffers.RecvBuffer(maxsize=self.options.max_msg_size)
    try:
      conn.recvbuf.recv(conn.sock)
    except ConnectionResetError:
      self._disconnect(conn)
      return
    if conn.recvbuf.done():
      if self.recvq.qsize() > self.options.max_recv_queue:
        raise RuntimeError('Too many incoming messages enqueued')
      self.recvq.put((conn.addr, conn.recvbuf.result()))
      conn.recvbuf = None

  def _disconnect(self, conn):
    self._log(f'Closed connection to {conn.addr[0]}:{conn.addr[1]}')
    conn = self.conns.pop(conn.addr)
    if conn.sendbufs:
      count = len(conn.sendbufs)
      conn.sendbufs.clear()
      self._log(f'Dropping {count} messages to disconnected client')
    self.sel.unregister(conn.sock)
    conn.sock.close()

  def _numsending(self):
    return sum(len(x.sendbufs) for x in self.conns.values())

  def _log(self, *args, **kwargs):
    if not self.options.logging:
      return
    contextlib.context.print(
        self.name, *args, color=self.options.logging_color)
